/*********************************************************
          Battery Management System
          Version Number V0.35 Pre-production
          Date 13.01.2023
 *********************************************************/
/*
  V0.35 Add Battery Temperature Restrictions 
  V0.34 Charging Control Reduction Added
  V0.22 Status Selection Added, plus Battery Balancing Corrected
  V0.21 Battery Balancing with a few controls
  V0.20 Basic Battery Balancing added without control
*/


// Include libraries
#include "SPI.h"
#include <Arduino.h>
//#include <mcp_can.h>
#include <mcp2515.h>
#include "config.h"
//#include "SerialConsole.h"
//#include "logger.h"
#include <EEPROM.h>

// Constant declaration
// status MASKs
const uint8_t statusClear = 0x00;
const uint8_t statusLowVoltage = 0x01;
const uint8_t statusVeryLowVoltage = 0x02;
const uint8_t statusHighVoltage = 0x04;
const uint8_t statusAllowDischarge = 0x08;
const uint8_t statusChargeFlag = 0x10;
const uint8_t statusAllowCharging = 0x20;
const uint8_t statusCharging = 0x40;
const uint8_t statusChargeComplete = 0x80;


//StatusToChargerLED
const uint8_t statusTCLEDClear = 0x00;
const uint8_t statusTCLEDLowVoltage = 0x01;
const uint8_t statusTCLEDVeryLowVoltage = 0x02;
const uint8_t statusTCLEDOVFault = 0x04;
const uint8_t statusTCLEDError = 0x08;
const uint8_t statusTCLEDChargeReady = 0x10;
const uint8_t statusTCLEDChargePaused = 0x20;
const uint8_t statusTCLEDCharging = 0x40;
const uint8_t statusTCLEDChargeComplete = 0x80;



byte Status = statusClear;
byte statusToChargerLED = statusTCLEDClear;


// errorByte MASK
const uint8_t initError = 0x01;
const uint8_t mismatchBefore = 0x02;
const uint8_t mismatchAfter = 0x04;
const uint8_t receiveBuffer = 0x08;

// Battery constants

//const uint32_t fullCapacity = 161021952000;  // 24.85 * 1024 * 1000 * 3600 / 25 = (43.68 * 147456000)
const int16_t highVoltageCutOff = 4175;      // Define high voltage cut off, in mV
const int16_t highVoltageFullCharge = 4100;  // Define high voltage fully charge
const int16_t lowVoltageWarning = 3600;      // Define low voltage warning, in mV
const int16_t lowVoltageCutOff = 3200;       // Define low voltage cut off, in mV
const int16_t voltageAllowance = 10;         // Define the accuracy the balancing algorithm can balance it to, in mV
const int16_t startShuntVoltage = 3800;      // Define start shunting voltage, in mV
const int16_t cellHysteresis = 100;          //Cell Hysteresis after Charging mV
long fullCapacity = 40000;                   // Wh
long startupCapacity;                        // To to loaded from EEPROM at start Wh
long currentCapacity;                        // same as startupCapacity at start Wh
int sessionUsage;                            // Wh
int startupCapacityAddress = 0;              // Set address for Capacity
int currentPower;                            // Power 1/10th KW



// Charging constants
const uint16_t maxChargeCurrent = 100;  // in dA (10.0A)
int targetSOC;
int BMSSOC;

// Other constants
const float chargeCompensationFactor = 0.988;     // Charge current compensation
const uint8_t balanceCellCounterTurnAround = 40;  // Every 10s (40 x 250ms)



// Variable declaration
uint16_t CellsEnabled = ALLCELLS;
uint16_t cellVoltageNew[MAX_MODULES][MAX_CELLS_PER_MODULE];
uint8_t NumModulesFound;

uint32_t watchdogPollTimer = millis();
uint32_t watchdogShutdownTimer;
uint32_t watchdogChargerPauseTimer;
uint32_t watchdogLowVoltagePauseTimer;
uint32_t watchdogVeryLowVoltagePauseTimer;
boolean moduleNTC3Select = false;

int tempHighCell;
int tempLowCell;
int tempHighDie;
boolean steadyReading = true;
int controlRegenPercent = 100;
int controlPowerPercent = 100;
int controlChargeRestriction = 100;
int controlTempHigh = 100;
int controlTempLow = 100;
int controlTempPercent = 100;

int RelayT = 0;  //Tremp

uint8_t cellBlockTemp[MAX_MODULES][SENSORS_PER_MODULE];  // 2 sensors per module
int16_t dieTemperature[MAX_MODULES];
uint32_t totalCapacity;
uint8_t dataReady;

uint16_t DebugStatus;
int incomingByte = 0;

StatusStoreD CellData;

EEpromsettings settings;
//SerialConsole console;
uint8_t TurnOnBalanceCells = 0;

// I/O-PINS
const uint8_t supplyLossPIN = 2;    // D2, Int0
const uint8_t chargeFlag = 4;       // D4 SHDN
const uint8_t SPI_MAX_CS_PIN = 10;  //ChipSelect PIN for MAX17841
const uint8_t SPI_MCP_CS_PIN = 8;   // ChipSelect PIN for CAN BUS
//long int Atest = 0;

// SPI-bus
// CLK, SCK - PIN 13
// MISO, SDO - PIN 12
// MOSI, SDI - PIN 11


// SPI setup
SPISettings MAX17841(4000000, MSBFIRST, SPI_MODE0);
//  CAN Bus setup
//MCP_CAN CAN(SPI_MCP_CS_PIN);  // Set CS pin
struct can_frame canMsg;
MCP2515 mcp2515(SPI_MCP_CS_PIN);  // chip select CS pin 8
void setup() {
  Serial.begin(115200);
  SPI.begin();
  pinMode(chargeFlag, INPUT_PULLUP);
  pinMode(supplyLossPIN, INPUT_PULLUP);

  pinMode(SPI_MAX_CS_PIN, OUTPUT);
  digitalWrite(SPI_MAX_CS_PIN, HIGH);  // No transaction

  pinMode(SPI_MCP_CS_PIN, OUTPUT);

  startupCapacity = (EEPROM.read(startupCapacityAddress) << 8) + (EEPROM.read(startupCapacityAddress + 1));

  currentCapacity = startupCapacity;



  Serial.print("Starting");
  for (int i = 0; i < 7; i++) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("*********************************************************");
  Serial.println("*        Battery Management System                      *");
  Serial.println("*        Version Number V0.35 Pre-production            *");
  Serial.println("*        Date 13.01.2023                                *");
  Serial.println("*********************************************************");

  Serial.println("");

  pinMode(3, OUTPUT);


  dataReady = 0x02;  // Reset data ready
  daisyChainInit();
  setupSlaves();
  Serial.println("all slaves balance off ");
  writeAllSlaves(0x1A, 0x0000, true);  // Turn off cell
  measureCellData();
  readData();
}

void loop() {
  // Main Looping Programme
  
  CapacityCalculate();

  if (millis() - watchdogPollTimer >= POLL_FREQUENCY) {
    measureCellData();
    readData();


    if (moduleNTC3Select == false) {
      moduleNTC3Select = true;
      writeAllSlaves(0x11, 0x7002, true);  // Set GPIO for NTC 3 Selection
    } else {
      moduleNTC3Select = false;
      writeAllSlaves(0x11, 0x7001, true);  // Set GPIO for NTC 3 Selection
    }

    FindHighLowCell();              // Find High and Low Voltages
    tempHighCell = highTempCell();  // Find High Temperature
    tempLowCell = lowTempCell();    // Find Low Temperature
    tempHighDie = highTempDie();    // Find High DIE Temperature

 

    /***Routine to select charging or stop depending on charge flap***
        watchdogChargerPauseTimer = 0 */
    if ((digitalRead(chargeFlag) == 0) && ((Status & statusChargeFlag) != statusChargeFlag)) {
      Status |= statusChargeFlag;
      watchdogChargerPauseTimer = 0;
      steadyReading = true;
    }
    if ((digitalRead(chargeFlag) == 1) && ((Status & statusChargeFlag) == statusChargeFlag)) {
      Status &= ~statusChargeFlag;
      watchdogChargerPauseTimer = 0;
      steadyReading = true;
    }
    
    balanceCells(CellData.MinVoltage.Voltage);
    if (((Status & statusChargeFlag) == statusChargeFlag) && ((Status & statusChargeComplete) != statusChargeComplete))  // If charge flag is open and Charge is not Complete start charging monitoring
    {
      updateBatteryChargingStatus();
      // controlChargeRestriction;
      Serial.print("Status: ");
      Serial.println(Status, BIN);
      Serial.print("High Cell: ");
      Serial.println(CellData.MaxVoltage.Voltage);
      steadyReading = balanceCells(CellData.MinVoltage.Voltage);
    }

    if ((Status & statusChargeFlag) != statusChargeFlag) {

      updateBatteryNotChargingStatus();
      steadyReading = true;
    }
    if (steadyReading == true) {
      Serial.print("High Cell: ");
      Serial.print(CellData.MaxVoltage.Voltage);
      Serial.print("V Module: ");
      Serial.print(CellData.MaxVoltage.Module);
      Serial.print(" Cell: ");
      Serial.print(CellData.MaxVoltage.CellNum);
      Serial.print("  Low Cell: ");
      Serial.print(CellData.MinVoltage.Voltage);
      Serial.print("V Module: ");
      Serial.print(CellData.MinVoltage.Module);
      Serial.print(" Cell: ");
      Serial.println(CellData.MinVoltage.CellNum);
      Serial.print("Total Voltage: ");
      Serial.println(CellData.TotalVoltage);
      Serial.print("High Cell Temp C: ");
      Serial.print(tempHighCell);
      Serial.print("  Low Cell Temp C: ");
      Serial.print(tempLowCell);
      Serial.print("  High Die Temp C: ");
      Serial.println(tempHighDie);
    }
    controlTempHigh = 100 - (pow((tempHighCell - 22), 4) * 0.0003);
    controlTempLow = 100- (pow((tempLowCell - 22), 4) * 0.0003);
    if (controlTempHigh < controlTempLow) {
      controlTempPercent = controlTempHigh;
Serial.println("High Temperature Restriction");      
    } else {
      controlTempPercent = controlTempLow;
      Serial.println("Low Temperature Restriction");
    }

    if (CellData.MaxVoltage.Voltage > (highVoltageCutOff - 200)) {
      controlRegenPercent = (highVoltageCutOff - CellData.MaxVoltage.Voltage) / 2;
    } else {
      controlRegenPercent = 100;
    }

    controlRegenPercent = constrain(controlRegenPercent, 0, controlTempPercent);

    if (CellData.MinVoltage.Voltage < (lowVoltageCutOff + 200)) {
      controlPowerPercent = (CellData.MinVoltage.Voltage - lowVoltageCutOff) / 2;
    } else {
      controlPowerPercent = 100;
    }
    controlPowerPercent = constrain(controlPowerPercent, 0, controlTempPercent);


    Serial.print("Power Control: ");
    Serial.println(controlPowerPercent);
    Serial.print("Regen Control: ");
    Serial.println(controlRegenPercent);

    Serial.print("Status: ");
    Serial.print(Status, BIN);
    Serial.print("   Charger LED Status: ");
    Serial.println(statusToChargerLED, BIN);

    if ((digitalRead(supplyLossPIN) == HIGH)) {
      Serial.println("Shutdown");
      Shutdown();
    }


    // controlChargeRestriction;
    watchdogPollTimer = millis();
  }
}
