

/*******************************************
  UART Daisy-Chain Initialization Sequence
*******************************************/
uint8_t daisyChainInit()
{
  uint8_t check;
  uint8_t data[4];
  uint8_t errorByte = 0x00;

  SPI.beginTransaction(MAX17841);

  // 1, Enable Keep-Alive mode
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x10); // Write Configuration 3 register
  SPI.transfer(0x05); // Set keep-alive period to 160μs
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 2, Enable Rx Interrupt flags for RX_Error and RX_Overflow
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x04); // Write RX_Interrupt_Enable register
  SPI.transfer(0x88); // Set the RX_Error_INT_Enable and RX_Overflow_INT_Enable bits
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 3, Clear receive buffer
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xE0); // Clear receive buffer
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 4, Wake-up UART slave devices
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x0E); // Write Configuration 2 register
  SPI.transfer(0x30); // Enable Transmit Preambles mode
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 5, Wait for all UART slave devices to wake up
  uint32_t watchdogTimer = millis();
  check = 0;
  while (check != 0x21) // If RX_Status = 21h, continue. Otherwise, repeat transaction until true or timeout
  {
    digitalWrite(SPI_MAX_CS_PIN, LOW);
    SPI.transfer(0x01); // Read RX_Status register
    check = SPI.transfer(0x01); // Read RX_Status register
    digitalWrite(SPI_MAX_CS_PIN, HIGH);

    if ((millis() - watchdogTimer) > 100) // Watchdog timer of 100ms
    {
      Serial.print("Check:");
      Serial.print(check, HEX);

      Serial.println("  UARTSlaveDevicesWakeUp WD timeout");
      break ;
    }
  }
  Serial.println("UARTSlaveDevicesWakeUp done 1");
  // 6, End of UART slave device wake-up period
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x0E); // Write Configuration 2 register
  SPI.transfer(0x10); // Disable Transmit Preambles mode
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 7, Wait for null message to be received
  // 8, Clear transmit buffer
  // 9, Clear receive buffer
  clearBuffers();

  Serial.println("UARTSlaveDevicesWakeUp done 2");

  // 10, Load the HELLOALL command sequence into the load queue
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC0); // WR_LD_Q SPI command byte (write the load queue)
  SPI.transfer(0x03); // Message length
  SPI.transfer(0x57); // HELLOALL command byte
  SPI.transfer(0x00); // Register address (0x00)
  SPI.transfer(0x00); // Initialization address of HELLOALL
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 11, Verify contents of the load queue
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC1); // RD_LD_Q SPI command byte

  for (int i = 0; i < 4; i++)
  {
    data[i] = SPI.transfer(0xC1);
  }

  if (!((data[0] == 0x03) && (data[1] == 0x57) && (data[2] == 0x00) && (data[3] == 0x00)))
  {
    errorByte |= initError;
    Serial.print("errorBytes1:");
    Serial.print(data[0], HEX);
    Serial.print(data[1], HEX);
    Serial.print(data[2], HEX);
    Serial.print(data[3], HEX);
  }
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 12, Transmit HELLOALL sequence
  // 13, Poll RX_Stop_Status bit
  transmitQueue();

  // 14, Read the HELLOALL message that propagated through the daisy-chain and was returned back to the ASCI
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x93); // RD_NXT_MSG SPI transaction

  for (int i = 0; i < 3; i++)
  {
    data[i] = SPI.transfer(0x93);
  }

  if (!((data[0] == 0x57) && (data[1] == 0x00) ))  //&& (data[2] == 0x02))) // Changed to 2 modules
  {
    //    errorByte |= initError;
    //    errorByte |= 2;
    Serial.print("errorBytes2:");
    Serial.print(data[0], HEX);
    Serial.print(data[1], HEX);
    Serial.print(data[2], HEX);

  }
  else
  {
    NumModulesFound = data[2];
    Serial.print("Modules:");
    Serial.println(NumModulesFound, DEC);
  }

  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 15, Check for receive buffer errors
  errorByte |= receiveBufferError();

  SPI.endTransaction();

  if (errorByte) // Error
  {
    Serial.println(errorByte, HEX);
    errorByte &= 0x00; // Clear errors
    Serial.println("errorByte cleared");
    //    daisyChainInit(); // Redo Daisy chain init do not do this it causes stack issues
  }
}
